/* This program is free software. It comes without any warranty, to
 * the extent permitted by applicable law. You can redistribute it
 * and/or modify it under the terms of the Do What The Fuck You Want
 * To Public License, Version 2, as published by Sam Hocevar. See
 * http://sam.zoy.org/wtfpl/COPYING for more details. */ 

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

//---------------------------------------------------------------------------------------------

// Global variables
int nLen;
char szLine[1024];
FILE *fz, *fk, *fr;

// Variables set for each romset
char gName[32];
char gNameb[32];
char year[5];
char parent[32];
char title[256];
char developer[128];
char gInit[32];

int disabled = 0;
int init = 0;
int controller_type = 0;
int larger_char_set = 0;
int screen_size_320 = 1;
unsigned char pcm_xor = 0;
unsigned char gfx_xor = 0;

unsigned short flags = 0;
// 0001 swapc
// 0002 swapp
// 0004 alternate text
// 0008 pcb
// 0010 hardware_encrypted_a
// 0020 hardware_encrypted_b
// 0040 p32
// 0100 working
// 0200 proto
// 0400 bootleg
// 0800 clone

// store romset information (rom names, CRCs, types, and sizes)
int gofst = 0;
char nName[32][64];
unsigned int nCRC[32];
unsigned int nSize[32];
unsigned char nType[32];

// Store init names (some clones use same as parent)
int iCnt = 0;
char init_names[512][32];

//---------------------------------------------------------------------------------------------

// take the information we've gathered and write it to a file
void write_driver()
{
	char types[10][24] = {
		{ "1 | BRF_ESS | BRF_PRG" },
		{ "2 | BRF_GRA" },
		{ "3 | BRF_GRA" },
		{ "4 | BRF_ESS | BRF_PRG" },
		{ "0 | BRF_PRG | BRF_OPT" },
		{ "5 | BRF_SND" },
		{ "6 | BRF_SND" },
		{ "0 | BRF_BIOS" },
		{ "" },
		{ "9 | BRF_ESS | BRF_PRG" }, // SMA
	};

	int i,j,k,l,cnt, r;
	int namelen = 0;
	int sizepad = 0;

	fprintf (fk, "// %s\n\n", title);
	fprintf (fk, "static struct BurnRomInfo %sRomDesc[] = {\n", gNameb);

	for (i = 0; i < gofst; i++)
	{
		if (strlen (nName[i]) > namelen && nType[i] != 4) namelen = strlen (nName[i]);
		if (nSize[i] >= 0x1000000) sizepad = 1;

		if (nSize[i] >= 0x1000000 && nType[i] == 3)
			flags |= 0x0008;

		if (nType[i] == 4 && nType[i+1] == 4) {
			for (j =0; j < (int)strlen(nName[i]); j++) {
				if (nName[i][j] == 'd' || nName[i][j] == 'D') {
					nType[i+1] = 5;
					break;
				}
				if (nName[i+1][j] == 'd' || nName[i+1][j] == 'D') {
					nType[i] = 5;
					break;
				}
			}
		}
	}
	if (namelen < 11)
		namelen += 11 - namelen;
	else
	{
		if (namelen < 13 && namelen != 11) namelen -= 13 - namelen;
	}

	namelen+=3;

	for (i = 1, l = 1, cnt = 0; i < 9; i++)
	{
		if (i == 8) continue; // skip bios

		for (j = 0, r = 0; j < gofst; j++)
		{
			if (nType[j] == i || (nType[j] == 10 && i == 1))
			{
				if (l == 4 && i == 5) l = 5;
				if (l == 6 && i == 7) l = 7;

				if (i != l) fprintf (fk, "\n");
				l = i;

				fprintf (fk, "\t{ \"%s\",", nName[j]);

				for (k = 0; k < namelen-1; k++)
				{
					if (k > strlen (nName[j]))
					{
						fprintf (fk, " ");
					}
				}

				fprintf (fk, "0x%6.6x,%s 0x%8.8x, %s },%s // %s%d ", nSize[j], ((sizepad && (nSize[j] < 0x1000000)) ? " " : ""), nCRC[j], types[nType[j]-1], ((strlen(types[nType[j]-1]) < 20) ? "          " : ""), (cnt < 10) ? " " : "", cnt);
				if (r == 0) {
					if (i == 1) fprintf (fk, "68K code");
					if (i == 2) fprintf (fk, "Text layer tiles");
					if (i == 3) fprintf (fk, "Sprite data");
					if (i == 4 || i == 5) fprintf (fk, "Z80 code");
					if (i == 4 && nType[j-1] == 5) fprintf (fk, " (decrypted)");
					if (i == 6) fprintf (fk, "Sound data");
				}
				fprintf (fk, "\n");
				cnt++;
				r=1;
			}
		}
	}

	fprintf (fk, "};\n\n");

	fprintf (fk, "STDROMPICKEXT(%s, %s, neogeo);\n", gNameb, gNameb);
	fprintf (fk, "STD_ROM_FN(%s);\n\n", gNameb);

	if (larger_char_set && !init) {
		memcpy (gInit, gName, strlen(gName));
		init = 1;
	}

	if (init & 1)
	{
		sprintf (szLine, "inits\\%s.txt", gInit);
		fr = fopen(szLine, "rt");
		if (fr != NULL)
		{
			while (1)
			{
				if (fgets (szLine, 1024, fr) == NULL)
					break;	
				fprintf (fk, "%s", szLine);
			}
			fclose (fr);
			if (szLine[strlen(szLine)-1] != '\n') fprintf (fk, "\n");
			if (szLine[strlen(szLine)  ] != '\n') fprintf (fk, "\n");
		}
		else
		{
			fprintf (fk, "int %sInit()\n", gInit);
			fprintf (fk, "{\n");

			if (pcm_xor >> 4) fprintf (fk, "\tint nRet;\n\n");
			if (pcm_xor & 0x10) fprintf (fk, "\textern unsigned char* YM2610ADPCMAROM;\n\n");

			if (larger_char_set) fprintf (fk, "\tnNeoTextROMSize = 0x%6.6x;\n", 0x40000 << larger_char_set);
			if (flags&0x0030) fprintf (fk, "\tnNeoProtectionXor = 0x%2.2x;\n", gfx_xor & 0xff);

			if (pcm_xor & 0x30)
			{
				fprintf (fk, "\n\tnRet = NeoInit();\n\n");
				fprintf (fk, "\tif (nRet == 0) {\n");

				if (pcm_xor & 0x20)
				{
					fprintf (fk, "\t\tPCM2DecryptVInfo Info = { ");

					switch (pcm_xor & 0x0f)
					{
						case 0x00:
							fprintf (fk, "0xA5000, 0x000000, { 0xF9, 0xE0, 0x5D, 0xF3, 0xEA, 0x92, 0xBE, 0xEF");
						break;
						case 0x01:
							fprintf (fk, "0x01000, 0xFFCE20, { 0xC4, 0x83, 0xA8, 0x5F, 0x21, 0x27, 0x64, 0xAF");
						break;
						case 0x02:
							fprintf (fk, "0x4E001, 0xFE2CF6, { 0xC3, 0xFD, 0x81, 0xAC, 0x6D, 0xE7, 0xBF, 0x9E");
						break;
						case 0x03:
							fprintf (fk, "0xC2000, 0xFFAC28, { 0xC3, 0xFD, 0x81, 0xAC, 0x6D, 0xE7, 0xBF, 0x9E");
						break;
						case 0x04:
							fprintf (fk, "0x0A000, 0xFEB2C0, { 0xCB, 0x29, 0x7D, 0x43, 0xD2, 0x3A, 0xC2, 0xB4");
						break;
						case 0x05: 
							fprintf (fk, "0xA7001, 0xFF14EA, { 0x4B, 0xA4, 0x63, 0x46, 0xF0, 0x91, 0xEA, 0x62");
						break;
						case 0x06:
							fprintf (fk, "0x02000, 0xFFB440, { 0x4B, 0xA4, 0x63, 0x46, 0xF0, 0x91, 0xEA, 0x62");
						break;
					}

					fprintf (fk, " } };\n\n");

					fprintf (fk, "\t\tPCM2DecryptV(&Info);\n");
				}

				if (pcm_xor & 0x10)
				{
					k = (pcm_xor & 0x0e) | ((pcm_xor & 1) << 4);
					l = 0;

					for (i = 0; i < gofst; i++)
					{
						if (nType[i] == 6) l+= nSize[i];
					}					

					fprintf (fk, "\t\tfor (int i = 0; i < 0x%x / 2; i += %d / 2) {\n", l, k);
					fprintf (fk, "\t\t\tunsigned short buffer[%d / 2];\n", k);
					fprintf (fk, "\t\t\tmemcpy(buffer, ((unsigned short*)YM2610ADPCMAROM) + i, %d);\n", k);
					fprintf (fk, "\t\t\tfor (int j = 0; j < %d / 2; j++) {\n", k);
					fprintf (fk, "\t\t\t\t((unsigned short*)YM2610ADPCMAROM)[i + j] = buffer[j ^ %d];\n", k / 4);
					fprintf (fk, "\t\t\t}\n");
					fprintf (fk, "\t\t}\n");
				}

				fprintf (fk, "\t}\n\n");

				fprintf (fk, "\treturn nRet;\n");
			} else
				fprintf (fk, "\treturn NeoInit();\n");
			fprintf (fk, "}\n\n");
		}
	}

	fprintf (fk, "struct BurnDriver%sBurnDrv%s = {\n", disabled ? "X " : " ", gNameb);
	fprintf (fk, "	\"%s\", ", gName);
	if (~flags&0x0800)fprintf (fk, "NULL");
	else fprintf (fk, "\"%s\"", parent);
	fprintf (fk, ", \"neogeo\", \"%s\",\n", year);
	fprintf (fk, "	\"%s\\0\", NULL, \"%s\", \"Neo Geo\",\n", title, developer);
	fprintf (fk, "	NULL, NULL, NULL, NULL,\n");

	fprintf (fk, "	");

	if (flags&0x0100) {
		fprintf (fk, "BDF_GAME_WORKING");
		if (flags&0x0e00) fprintf (fk, " | ");
	}
	if (flags&0x0800) {
		fprintf (fk, "BDF_CLONE");
		if (flags&0x0600) fprintf (fk, " | ");
	}
	if (flags&0x0400) {
		fprintf (fk, "BDF_BOOTLEG");
		if (flags&0x0200) fprintf (fk, " | ");
	}
	if (flags&0x0200) fprintf (fk, "BDF_PROTOTYPE");
	if ((flags&0x0f00) == 0x0000) fprintf (fk, "0");

	fprintf (fk, ", 2, HARDWARE_SNK_NEOGEO");

	if (flags&0x0002) fprintf (fk, " | HARDWARE_SNK_SWAPP");
	if (flags&0x0001) fprintf (fk, " | HARDWARE_SNK_SWAPC");
	if ((flags&0x0004) && larger_char_set) fprintf (fk, " | HARDWARE_SNK_ALTERNATE_TEXT");

	switch (controller_type)
	{
		case 2:
			fprintf (fk, " | HARDWARE_SNK_MAHJONG");
		break;

		case 3:
			fprintf (fk, " | HARDWARE_SNK_TRACKBALL");
		break;

		case 4:
			fprintf (fk, " | HARDWARE_SNK_PADDLE"); 
		break;

		case 5:
		case 6:
			fprintf (fk, " | HARDWARE_SNK_GAMBLING");
		break;
	}

	if (nType[0] == 10) fprintf (fk, " | HARDWARE_SNK_SMA_PROTECTION");

	switch ((flags>>4)&3)
	{
		case 0: break;
		case 1:
			fprintf (fk, " | HARDWARE_SNK_ENCRYPTED_A");
		break;
		case 2:
			fprintf (fk, " | HARDWARE_SNK_ENCRYPTED_B");
		break;
	}

	if (flags&0x0008 && flags&0x0030) fprintf (fk, " | HARDWARE_SNK_PCB");

	if (nType[0] == 1 && nSize[0] == 0x400000 && nType[1] == 1 && nSize[1] == 0x400000 && (flags & 0x0040))
		fprintf (fk, " | HARDWARE_SNK_P32");

	fprintf (fk, ",\n");
	fprintf (fk, "	NULL, %sRomInfo, %sRomName, ", gNameb, gNameb);

	switch (controller_type)
	{
		case 1:
		case 6:
			fprintf (fk, "neogeo");
		break;

		case 2:
			fprintf (fk, "neomahjong");
		break;

		case 3:
			fprintf (fk, "neotrackball");
		break;

		case 4:
			fprintf (fk, "neopaddle");
		break;

		case 5:
			fprintf (fk, "vliner");
		break;
	}

	fprintf (fk, "InputInfo, neogeoDIPInfo,\n");
	fprintf (fk, "	%sInit, NeoExit, NeoFrame, NeoRender, NeoScan, &NeoRecalcPalette,\n", (init) ? gInit : "Neo");
	fprintf (fk, "	%d, 224, 4, 3\n", (screen_size_320) ? 320 : 304);
	fprintf (fk, "};\n\n\n");
}


//---------------------------------------------------------------------------------------------


// find the commas, ignoring those between quotes
int revlen(char *a, int o)
{
	int i, q;
	for (i = o-1, q = 0; i > 0; i--)
	{
		if (a[i] == '\"') q ^= 1;	// Quote
		if ((a[i] == ' ' || a[i] == ',') && q != 1)
			return i;
	}

	return o;
}

// convert individual chars to dec/hex
unsigned char char_to_dec(unsigned char a)
{
	if ((a & 0xf0) == 0x30 && (a & 0x0f) <= 0x09) return a & 0x0f;		// 0 - 9
	if ((a & 0xd0) == 0x40 && (a & 0x0f) <= 0x06) return (a & 0x07) + 9;	// a - f

	return 0xff; // invalid
}

// go through the list of inits that have already been used
int get_inits()
{
	int i;
	if (strcmp ("neogeo", gInit) == 0)	return 0;	// don't show init or use init name

	for (i = 0; i < iCnt; i++) {
		if (strcmp (init_names[i], gInit) == 0) { // already in the list
			return 2;				// don't show init, but use init name
		}
	}

	strcpy (init_names[iCnt], gInit); // not found, add it to the list
	iCnt++;

	return 1;						// show init and init name
}


// find our init, go through it and find various info
void find_init()
{
	int i, ifound = 0;

	while (1)
	{
		if (fgets (szLine, 1024, fz) == NULL)
			break;

		nLen = strlen (szLine);

		if (strncmp ("GAME(", szLine, 5) == 0) break;

		if (!ifound)
		{
			for (i = 0; i < nLen; i++)
				if (strncmp ("DRIVER_INIT", szLine + i, 11) == 0) break;

			if (i == nLen) continue;

			for (i = 0; i < nLen; i++)
			{
				if (strncmp (szLine + i, gInit, strlen(gInit)) == 0 && szLine[i-1] == ' ')
				{
					ifound = 1;
					break;
				}
			}

			continue;
		}

		if (strncmp (szLine, "}", 1) == 0) break;

		for (i = 0; i < nLen; i++)
		{
			if (!strncmp (szLine + i, "kof2003_decrypt_68k();",  22)) flags |= 0x0040;
			if (!strncmp (szLine + i, "kf2k3pcb_decrypt_68k();", 23)) flags |= 0x0040;
			if (!strncmp (szLine + i, "mslug5_decrypt_68k();",   20)) flags |= 0x0040;
			if (!strncmp (szLine + i, "svcchaos_px_decrypt();",  22)) flags |= 0x0040;
			if (!strncmp (szLine + i, "kof2003_decrypt_68k();",  22)) flags |= 0x0040;

			if (strncmp (szLine + i, "neogeo_fixed_layer_bank_type = 2;", 33) == 0)
			{
				flags |= 0x0004;
			}

			if (strncmp (szLine + i, "kof99_neogeo_gfx_decrypt(", 25) == 0)
			{
				flags |= 0x0010;
				if (szLine[i + 25] == ' ') i++;
				gfx_xor |= char_to_dec(szLine[i + 27]) << 4;
				gfx_xor |= char_to_dec(szLine[i + 28]);
				break;
			}

			if (strncmp (szLine + i, "kof2000_neogeo_gfx_decrypt(", 27) == 0)
			{
				flags |= 0x0020;
				if (szLine[i + 27] == ' ') i++;
				gfx_xor |= char_to_dec(szLine[i + 29]) << 4;
				gfx_xor |= char_to_dec(szLine[i + 30]);
				break;
			}

			if (strncmp (szLine + i, "cmc50_neogeo_gfx_decrypt(", 25) == 0)
			{
				flags |= 0x0020;
				if (szLine[i + 25] == ' ') i++;
				gfx_xor |= char_to_dec(szLine[i + 27]) << 4;
				gfx_xor |= char_to_dec(szLine[i + 28]);
				break;
			}
	
			if (strncmp (szLine + i, "neo_pcm2_snk_1999(", 18) == 0)
			{
				pcm_xor = 0x10;
				if (szLine[i + 18] == ' ') i++;
				pcm_xor |= char_to_dec(szLine[i + 18]) & 0x0f;
				break;
			}

			if (strncmp (szLine + i, "neo_pcm2_swap(", 14) == 0)
			{
				pcm_xor = 0x20;
				if (szLine[i + 14] == ' ') i++;
				pcm_xor |= char_to_dec(szLine[i + 14]) & 0x0f;
				break;
			}
		}
	}
}

// find the 'GAME(' line and then strip out various information
void game_info()
{

#define cpy(dst, o)	\
	while (szLine[c[o]-1] == ' ') c[o]--;	\
	ofst = revlen(szLine, c[o]);	\
	memcpy (dst, szLine + ofst + 1, c[o] - ofst - 1);	\
	dst[c[o] - ofst - 1] = '\0';	\

#define remquote(dst)	\
	if (dst[0] == '\"') memcpy (dst, dst + 1, strlen(dst) + 1);	\
	if (dst[strlen(dst)-1] == '\"') dst[strlen(dst)-1] = '\0'	\

	char tmp[256];
	int i, c[16], c0, q0 = 0, ofst;

	while (1)
	{
		if (fgets (szLine, 1024, fz) == NULL)
			break;

		nLen = strlen (szLine);

		if (nLen < 6 || szLine[0] == '/' || strncmp ("GAME", szLine, 4)) continue;

		for (i = 0, c0 = 0; i < nLen && c0 < 16; i++)
		{
			if (szLine[i] == '\"') q0 ^= 1;			// Quote
			if (szLine[i] == ',' && q0 == 0) c[c0++] = i;
		}
	
		cpy(tmp,	1);
		if (strcmp (gName, tmp)) continue;

		cpy(year,	0);

		cpy(parent,	2);
		flags |= strcmp (parent, "neogeo") ? 0x0800: 0;
		if (strcmp (parent, "0") == 0) flags &= 0xf7ff;
	
		cpy(tmp,	3);	// screen size
		screen_size_320 = strncmp ("neogeo_s", tmp, 8) ? 1 : 0;
	
		cpy(tmp,	4);	// input

		controller_type = 1;					// neogeo
		if (strcmp ("mjneogeo", tmp) == 0) controller_type = 2;	// mahjong
		if (strcmp ("irrmaze",  tmp) == 0) controller_type = 3;	// trackball
		if (strcmp ("popbounc", tmp) == 0) controller_type = 4; // popbounc
		if (strcmp ("vliner",   tmp) == 0) controller_type = 5; // vliner
		if (strcmp ("jockeygp", tmp) == 0) controller_type = 6;	// jockeygp
	
		cpy(gInit,	5);
		init = get_inits();

		cpy(developer,	7);
		remquote(developer);

		cpy(title,	8);
		remquote(title);

		for (i = 0; i < strlen (title); i++)
		{
			if (strncmp (title + i, "bootleg", 7) == 0) {
				flags |= 0x0400;
				break;
			}
	
			if (strncmp (title + i, "proto", 5) == 0) {
				flags |= 0x0200;
				break;
			}
		}
	
		memcpy (tmp, szLine + c[c0-1] + 2, strlen(szLine) - (c[c0-1] + 2));
		flags |= strncmp ("GAME_NOT_WORKING", tmp, 16) ? 0x0100 : 0;
	}
}

// go through the lines that contain actual ROM info (name, crc, length)
void strip_rom()
{
	int q0 = 0, o1 = 0;
	char *tmp;

	int i, j, k;
	for (i = 0; i < nLen; i++)
	{
		if (!q0 && szLine[i] == '\"')
		{
			q0 = i;
			continue;
		}

		if (q0 && szLine[i] == '\"')
		{
			memcpy (nName[gofst], szLine + q0 + 1, i - (q0 + 1));
			nName[gofst][i-(q0 + 1)] = '\0';
			continue;
		}

		if (!o1 && szLine[i] == '0' && szLine[i+1] == 'x')
		{
			o1 = i;
			continue;
		}

		if (o1 && szLine[i] == '0' && szLine[i+1] == 'x')
		{
			nSize[gofst] = 0;
			tmp = szLine + i + 2;

			for (j = 6, k = 0; j >= 0; j--)
			{
				if (char_to_dec(tmp[j]) != 0xff)
				{
					nSize[gofst] |= char_to_dec(tmp[j]) << (k * 4);
					k++;
				}

				i++;
			}

			continue;
		}

		if (strncmp ("CRC(", szLine + i, 4) == 0)
		{
			nCRC[gofst] = 0;
			tmp = szLine + i + 4;

			for (j = 8, k = 0; j >= 0; j--)
			{
				if (char_to_dec(tmp[j]) != 0xff)
				{
					nCRC[gofst] |= char_to_dec(tmp[j]) << (k * 4);
					k++;
				}

				i++;
			}

			continue;
		}
	}
}

// we found a ROM_CONTINUE, so add the size it is continuing to the rom from the previous line
void do_rom_continue()
{
	int i, j, k, f;
	unsigned int add = 0;
	for (i = 0, f = 0, j = 0; i < nLen; i++) {
		if (szLine[i] == '0' && szLine[i+1] == 'x') {
			if (!f) {
				f = 1;
				continue;
			} else {
				for (j = 8, k = 0; j >= 0; j--) {
					if (char_to_dec(szLine[i+j]) != 0xff) {
						add |= char_to_dec(szLine[i+j]) << (k * 4);
						k++;
					}
				}

				nSize[gofst-1] += add;
				break;
			}
		}
 	}
}

// structs that FBA uses for game names don't like numbers at the beginning of their names
// so move those to the end. (2020bb -> bb2020, 3countb -> countb3)
void do_gnameb()
{
	memcpy (gNameb, gName, strlen(gName)+1);

	int i;
	while (1)
	{
		if (gNameb[0] < 0x30 || gNameb[0] > 0x39) break;

		unsigned char t = gNameb[0];
		memcpy (gNameb, gNameb + 1, strlen(gNameb)-1);
		gNameb[strlen(gNameb)-1] = t;
	}
}

#define find_region(a, b) \
	if (strncmp (a, szLine + i, strlen(a)) == 0) {	\
		bType = b;	\
		break;		\
	}			\


//---------------------------------------------------------------------------------------------

int main()
{
	int i;
	int bType;
	int cnt = 0;
	int skip = 0;
	fpos_t posa, posb, posc;
	int processing_driver = 0;
	int apos = 0, bpos = 0, cpos = 0;

	memset (init_names, 0, 512 * 32);

	fz = fopen("neodrvr.c", "rt");

	if (fz == NULL)
	{
		printf ("neodrvr.c not found :(\n");
		return 1;
	}

	fk = fopen("out.txt", "wt");

	// file read offsets, 10x speedup because of this. :)
	while (1)
	{
		if (fgets (szLine, 1024, fz) == NULL)
			break;

		if (!apos) {	// find positions of the ROM definitions
			if (strncmp (szLine, "ROM_START(", 10) == 0) apos = 1;
			else fgetpos (fz, &posa);
		}

		if (!bpos) {	// find positions of the driver inits
			if (strncmp (szLine, "static DRIVER_INIT(", 19) == 0) bpos = 1;
			else fgetpos (fz, &posb);
		}

		if (!cpos) {	// find positions of the GAME infos
			if (strncmp (szLine, "GAME(", 5) == 0) {
				cpos = 1;
				break;
			}
			else fgetpos (fz, &posc);
		}
	}

	fsetpos (fz, &posa);

	while (1)
	{
		if (fgets (szLine, 1024, fz) == NULL)
			break;

		szLine[strlen( szLine )] = '\0'; // remove break;
		nLen = strlen( szLine );

		if (szLine[0] == '/') continue;

		if (strncmp ("ROM_START(", szLine, 10) == 0)
		{
			gfx_xor = 0;
			pcm_xor = 0;
			gofst = 0;
			flags = 0;
			init = 0;
			screen_size_320 = 1;
			larger_char_set = 0;

			for (i = 12; i < nLen; i++)
				if (szLine[i] == ' ') break;

			strncpy (gName, szLine + 11, i);
			gName[i-11] = '\0';

			if (strcmp (gName, "neogeo") == 0) continue;

			cnt++;

			processing_driver = 1;
			continue;
		}

		if (processing_driver)
		{
			if (szLine[0] == '/' && szLine[1] == '/')
				continue;

			if (strncmp ("#endif", szLine, 6) == 0) {
				skip = 0;
				continue;
			}

			if (strncmp ("#if 0", szLine, 5) == 0 || skip) {
				skip = 1;
				continue;
			}

			for (i = 0; i < nLen; i++)
			{
				if (szLine[i+0] == '/' && szLine[i+1] == '/')
					break;

				find_region("NEOGEO_REGION_MAIN_CPU_CARTRIDGE", 1);

				if (strncmp ("NEOGEO_REGION_FIXED_LAYER_CARTRIDGE", szLine + i, 35) == 0)
				{
					if (strncmp ("80000", szLine + (i - 7), 5) == 0)
						larger_char_set = 1;
					if (strncmp ("00000", szLine + (i - 7), 5) == 0)
						larger_char_set = 2;

					bType = 2;
					break;
				}

				find_region("NEOGEO_REGION_SPRITES", 3);
				find_region("NEOGEO_REGION_AUDIO_CPU_CARTRIDGE", 4);
				find_region("NEOGEO_REGION_AUDIO_CPU_ENCRYPTED", 5);
				find_region("NEOGEO_REGION_AUDIO_DATA_1", 6);
				find_region("NEOGEO_REGION_AUDIO_DATA_2", 7);
				find_region("NEOGEO_REGION_MAIN_CPU_BIOS", 8);

				if (strncmp ("ROM_CONTINUE", szLine + i, 12) == 0)
				{
					if (bType == 1) flags |= 0x0002;
					if (bType == 3) flags |= 0x0001;
					do_rom_continue();
					break;
				}

				if (strncmp ("NEO_SFIX_", szLine + i, 9) == 0)
				{
					nType[gofst] = 2;
					strip_rom();
					if (strncmp ("1024", szLine + i + 9, 4) == 0) nSize[gofst] = 0x100000;
					if (strncmp ("512",  szLine + i + 9, 3) == 0) nSize[gofst] = 0x080000;
					if (strncmp ("256",  szLine + i + 9, 3) == 0) nSize[gofst] = 0x040000;
					if (strncmp ("128",  szLine + i + 9, 3) == 0) nSize[gofst] = 0x020000;
					if (strncmp ("64",   szLine + i + 9, 2) == 0) nSize[gofst] = 0x010000;
					if (strncmp ("32",   szLine + i + 9, 2) == 0) nSize[gofst] = 0x008000;

					gofst++;
					break;
				}

				if (strncmp ("NEO_BIOS_AUDIO_", szLine + i, 15) == 0)
				{
					nType[gofst] = 4;
					strip_rom();
					if (strncmp ("1024", szLine + i + 15, 4) == 0) nSize[gofst] = 0x100000;
					if (strncmp ("512",  szLine + i + 15, 3) == 0) nSize[gofst] = 0x080000;
					if (strncmp ("256",  szLine + i + 15, 3) == 0) nSize[gofst] = 0x040000;
					if (strncmp ("128",  szLine + i + 15, 3) == 0) nSize[gofst] = 0x020000;
					if (strncmp ("64",   szLine + i + 15, 2) == 0) nSize[gofst] = 0x010000;
					if (strncmp ("32",   szLine + i + 15, 2) == 0) nSize[gofst] = 0x008000;

					gofst++;
					break;
				}

				if (strncmp ("ROM_LOAD", szLine + i, 8) == 0)
				{
					nType[gofst] = bType;
					strip_rom();
					if (strcmp("sfix.sfx", nName[gofst]) == 0) continue;
					if (strcmp("sm1.sm1",  nName[gofst]) == 0) continue;

					if (gofst == 0)
					{
						for (i = 0; i < strlen(nName[gofst]); i++)
						{
							if (strncmp ("-sma", nName[gofst] + i, 4) == 0) {
								nType[gofst] = 10;
								break;
							}
						}
					}

					gofst++;
					break;
				}
			}

			if (strncmp ("ROM_END", szLine, 7) == 0)
			{
				processing_driver = 0;

				fgetpos (fz, &posa);
				fsetpos (fz, &posc);

				game_info();

				fsetpos (fz, &posb);

				find_init();

				fsetpos (fz, &posa);

				do_gnameb();

				write_driver();

				printf ("%s\n", gName);

				continue;
			}
		}
		else
		{
			if (strncmp ("#if 0",  szLine, 5) == 0) disabled = 1;
			if (strncmp ("#endif", szLine, 6) == 0)	disabled = 0;
		}
	}

	printf ("%d drivers processed\n", cnt);

	fclose (fz);
	fclose (fk);
}
